# Generated by Django 3.2.7 on 2022-03-03 09:58

from django.db import migrations, models
import django.db.models.deletion
import time_stamped_model.models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0004_customreport'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='kanbanreportlane',
            options={'ordering': ('order',)},
        ),
        migrations.RemoveField(
            model_name='kanbanreportlane',
            name='description',
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='background_colour_field',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='heading_colour_field',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='multiple_end_period',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='multiple_start_period',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='multiple_type',
            field=models.PositiveIntegerField(choices=[(0, 'N/A'), (1, 'Daily (single date)'), (2, 'Daily (within two date)'), (3, 'Weekly (single date)'), (4, 'Weekly (within two date)'), (5, 'Monthly (single date)'), (6, 'Monthly (within two date)')], default=0),
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='multiple_type_date_field',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='multiple_type_end_date_field',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='multiple_type_label',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
        migrations.CreateModel(
            name='KanbanReportDescription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField(blank=True, null=True)),
                ('order', models.PositiveSmallIntegerField()),
                ('kanban_report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.kanbanreport')),
                ('report_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype')),
            ],
            options={
                'ordering': ('order',),
            },
        ),
        migrations.AddField(
            model_name='kanbanreportlane',
            name='kanban_report_description',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.kanbanreportdescription'),
        ),
    ]



import re

from django.apps import apps

from advanced_report_builder.field_utils import ReportBuilderFieldUtils
from advanced_report_builder.utils import get_report_builder_class


class DataMergeUtils(ReportBuilderFieldUtils):
    def get_menu_fields(self, base_model, report_builder_class,
                        menus=None, codes=None, code_prefix='', next_code_prefix='__',
                        previous_base_model=None, table=None):

        for report_builder_field in report_builder_class.fields:
            django_field, col_type_override, columns, _ = self.get_field_details(
                base_model=base_model,
                field=report_builder_field,
                report_builder_class=report_builder_class,
                table=table)
            for column in columns:
                full_id = code_prefix + column.column_name
                if menus is not None:
                    menus.append({'code': full_id, 'text': column.title})
                if codes is not None:
                    codes.add(full_id)

        for include_field, include in report_builder_class.includes.items():
            app_label, model, report_builder_fields_str = include['model'].split('.')

            new_model = apps.get_model(app_label, model)
            if new_model != previous_base_model:
                new_report_builder_class = get_report_builder_class(model=new_model,
                                                                    class_name=report_builder_fields_str)

                title = include.get('title')
                if title is None or title == '':
                    title = new_report_builder_class.title
                if menus is not None:
                    menu = []
                else:
                    menu = None
                self.get_menu_fields(base_model=new_model,
                                     report_builder_class=new_report_builder_class,
                                     menus=menu,
                                     codes=codes,
                                     code_prefix=f"{code_prefix}{include_field}{next_code_prefix}",
                                     next_code_prefix=next_code_prefix,
                                     previous_base_model=base_model,
                                     table=table)
                if menus is not None:
                    menus.append({'text': title, 'menu': menu})

    @staticmethod
    def get_data_merge_variables(html):
        variables = re.findall('{{\s*([^*\s*}}]+)\s*}}', html)

        all_fields = set()
        for variable in variables:
            if '|' in variable:
                field = variable.split('|')[0]
            elif '&' in variable:
                field = variable.split('&')[0]
            else:
                field = variable
            all_fields.add(field)

        variables = re.findall('{%\s*(if|elif|with)\s([^%}]+)\s*%}', html)

        for variable in variables:
            for field in variable[1].split(' '):
                if field != '' and (field == 'as' or field[0] in ['=', '<', '>']):
                    break
                if field not in ['', 'not', 'and' 'or'] and field[0] not in ['(', ')', '"', "'"]:
                    all_fields.add(field)
        return all_fields

    def get_data_merge_columns(self, base_model, report_builder_class, html, table):
        all_fields = self.get_data_merge_variables(html)
        columns = set()
        column_map = {}
        for field in all_fields:
            django_field, col_type_override, _, _ = self.get_field_details(base_model=base_model,
                                                                           field=field,
                                                                           report_builder_class=report_builder_class,
                                                                           table=table)
            if django_field is not None or isinstance(col_type_override.field, (list, tuple)):
                if field not in columns and f'.{field}' not in columns:
                    columns.add('.' + field)

                field_parts = field.split('__')

                if col_type_override is not None and field_parts[-1] != col_type_override.field:
                    column_map[field] = col_type_override.field
        return columns, column_map



# Generated by Django 3.2.7 on 2023-07-18 16:42

from django.db import migrations
import django_modals.model_fields.colour


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0011_auto_20230620_1449'),
    ]

    operations = [
        migrations.AlterField(
            model_name='singlevaluereport',
            name='tile_colour',
            field=django_modals.model_fields.colour.ColourField(blank=True, max_length=10, null=True),
        ),
        migrations.AlterField(
            model_name='target',
            name='colour',
            field=django_modals.model_fields.colour.ColourField(blank=True, help_text='The colour when it gets displayed on a report', max_length=10, null=True),
        ),
    ]



import base64
import json

from crispy_forms.layout import Div
from django.forms import CharField, ChoiceField, BooleanField
from django.urls import reverse
from django_menus.menu import MenuItem
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.modals import FormModal
from django_modals.processes import PROCESS_EDIT_DELETE, PERMISSION_OFF
from django_modals.widgets.colour_picker import ColourPickerWidget
from django_modals.widgets.select2 import Select2Multiple

from advanced_report_builder.models import FunnelChartReport
from advanced_report_builder.toggle import RBToggle
from advanced_report_builder.utils import split_attr, encode_attribute, decode_attribute, get_report_builder_class
from advanced_report_builder.views.charts_base import ChartBaseView, ChartBaseFieldForm
from advanced_report_builder.views.modals_base import QueryBuilderModalBaseMixin, QueryBuilderModalBase


class FunnelChartView(ChartBaseView):
    use_annotations = False

    template_name = 'advanced_report_builder/charts/funnel/report.html'

    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.chart_report = self.report.funnelchartreport
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.table.funnel_chart_report = self.chart_report
        self.table.datatable_template = 'advanced_report_builder/charts/funnel/middle.html'
        context['funnel_chart_report'] = self.chart_report
        return context

    def set_extra_number_field_kwargs(self, data_attr, options, multiple_index):
        funnel_colour = data_attr.get('funnel_colour') or '801C70'
        funnel_colour = self.add_colour_offset(funnel_colour, multiple_index=multiple_index)
        options.update({'colour': funnel_colour})

    def edit_report_menu(self, request, chart_report_id, slug_str):
        return [MenuItem(f'advanced_report_builder:funnel_chart_modal,pk-{chart_report_id}{slug_str}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary']),
                *self.duplicate_menu(request=self.request, report_id=chart_report_id)]

    def get_date_field(self, index, fields, base_model, table):
        return None


class FunnelChartModal(QueryBuilderModalBase):
    template_name = 'advanced_report_builder/charts/modal.html'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = FunnelChartReport
    widgets = {'report_tags': Select2Multiple}

    form_fields = ['name',
                   'notes',
                   'report_type',
                   'report_tags',
                   'axis_value_type',
                   'fields',
                   ]

    def form_setup(self, form, *_args, **_kwargs):

        self.add_query_data(form)

        url = reverse('advanced_report_builder:funnel_chart_field_modal',
                      kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'})

        return ('name',
                'notes',
                'report_type',
                'report_tags',
                'axis_value_type',
                FieldEx('fields',
                        template='advanced_report_builder/select_column.html',
                        extra_context={'select_column_url': url,
                                       'command_prefix': ''}),
                FieldEx('query_data',
                        template='advanced_report_builder/query_builder.html'),
                )


class FunnelChartFieldForm(ChartBaseFieldForm):

    def setup_modal(self, *args, **kwargs):
        data = json.loads(base64.b64decode(self.slug['data']))
        report_type, base_model = self.get_report_type_details()

        data_attr = split_attr(data)

        self.fields['title'] = CharField(initial=data['title'])

        self.fields['funnel_colour'] = CharField(required=False, widget=ColourPickerWidget)
        self.fields['funnel_colour'].initial = data_attr.get('funnel_colour')

        self.fields['has_filter'] = BooleanField(required=False, widget=RBToggle())
        self.fields['filter'] = CharField(required=False)

        if data_attr.get('has_filter') == '1':
            self.fields['has_filter'].initial = True
            if 'filter' in data_attr:
                self.fields['filter'].initial = decode_attribute(data_attr['filter'])

        self.fields['multiple_columns'] = BooleanField(required=False, widget=RBToggle())

        report_builder_class = get_report_builder_class(model=base_model,
                                                        report_type=report_type)

        fields = []
        self._get_query_builder_foreign_key_fields(base_model=base_model,
                                                   report_builder_class=report_builder_class,
                                                   fields=fields)

        self.fields['multiple_column_field'] = ChoiceField(choices=fields, required=False)

        if data_attr.get('multiple_columns') == '1':
            self.fields['multiple_columns'].initial = True
            self.fields['multiple_column_field'].initial = data_attr.get('multiple_column_field')

        super().setup_modal(*args, **kwargs)

    def get_additional_attributes(self):
        attributes = []
        self.get_report_type_details()

        attributes.append(f'funnel_colour-{self.cleaned_data["funnel_colour"]}')

        if self.cleaned_data['has_filter']:
            attributes.append('has_filter-1')

            if self.cleaned_data['filter']:
                b64_filter = encode_attribute(self.cleaned_data['filter'])
                attributes.append(f'filter-{b64_filter}')
            if self.cleaned_data['multiple_columns']:
                attributes.append('multiple_columns-1')
                attributes.append(f'multiple_column_field-{self.cleaned_data["multiple_column_field"]}')

        if attributes:
            return '-'.join(attributes)
        return None


class FunnelChartFieldModal(QueryBuilderModalBaseMixin, FormModal):
    form_class = FunnelChartFieldForm
    size = 'xl'
    template_name = 'advanced_report_builder/charts/modal_field.html'
    no_header_x = True
    helper_class = HorizontalNoEnterHelper

    @property
    def modal_title(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        return f'Edit {data["title"]}'

    def form_valid(self, form):
        selector = self.slug['selector']

        _attr = form.get_additional_attributes()
        self.add_command({'function': 'set_attr',
                          'selector': f'#{selector}',
                          'attr': 'data-attr',
                          'val': _attr})

        self.add_command({'function': 'html', 'selector': f'#{selector} span', 'html': form.cleaned_data['title']})
        self.add_command({'function': 'save_query_builder_id_query_data'})
        self.add_command({'function': 'update_selection'})
        return self.command_response('close')

    # noinspection PyMethodMayBeStatic
    def form_setup(self, form, *_args, **_kwargs):
        form.add_trigger('has_filter', 'onchange', [
            {'selector': '#filter_fields_div', 'values': {'checked': 'show'}, 'default': 'hide'}])

        form.add_trigger('multiple_columns', 'onchange', [
            {'selector': '#multiple_columns_fields_div', 'values': {'checked': 'show'}, 'default': 'hide'},
        ])

        return ['title',
                'funnel_colour',
                Div(FieldEx('has_filter',
                            template='django_modals/fields/label_checkbox.html',
                            field_class='col-6 input-group-sm'),
                    Div(
                        FieldEx('filter',
                                template='advanced_report_builder/datatables/fields/single_query_builder.html'),
                        FieldEx('multiple_columns',
                                template='django_modals/fields/label_checkbox.html',
                                field_class='col-6 input-group-sm'),
                        Div(
                            FieldEx('multiple_column_field'),
                            css_id='multiple_columns_fields_div'),
                        css_id='filter_fields_div'),
                    css_id='annotations_fields_div')
                ]

    def ajax_get_query_builder_fields(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']

        report_type_id = self.slug['report_type_id']
        query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)

        return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))



from calendar import monthrange

from django.utils.dates import MONTHS


def get_target_value(min_date, max_date, target, month_range=False):
    override_data = target.get_override_data()

    # We could find the proportion of the start month and end_month
    # we are interested in and multiply the target by the result.

    if min_date and max_date and override_data:
        start_month = min_date.month
        end_month = max_date.month
        total_target_value = 0

        if month_range:
            # if we know that the range is exactly a month long then we can just look it up in the dict.
            for year in range(min_date.year, max_date.year + 1):
                year_data = override_data.get(str(year))

                if year_data:
                    month_str = MONTHS.get(min_date.month)
                    return year_data.get(month_str, target.get_value())
                else:
                    return target.get_value()
        else:
            # if it isn't exactly a month long, then we need to do some calculations to find the exact amount.

            for year in range(min_date.year, max_date.year + 1):
                year_data = override_data.get(str(year))
                yearly_target_value = 0

                for month in range(start_month, end_month + 1):
                    if year_data:
                        month_str = MONTHS.get(month)
                        month_value = year_data.get(month_str)
                    else:
                        month_value = target.get_value()

                    days_in_month = monthrange(year, month)[1]

                    # Work out the days inside the target in this month:
                    if (month == min_date.month and year == min_date.year) and \
                            (month == max_date.month and year == max_date.year):
                        # if start and end dates are in the same month and year, just subtract min from max.
                        days_in_target = (max_date - min_date).days
                    elif month == min_date.month and year == min_date.year:
                        days_in_target = (days_in_month - min_date.day) + 1
                    elif month == max_date.month and year == max_date.year:
                        days_in_target = max_date.day
                    else:
                        days_in_target = days_in_month

                    # The actual target value for this month will have to be a proportion.
                    # (Since the start date might be after the 1st of the month
                    # and the end_date might be before the last day of the month.
                    target_in_month = (days_in_target / days_in_month) * month_value
                    yearly_target_value += target_in_month

                total_target_value += yearly_target_value

        return total_target_value
    else:
        # we don't have any dates to compare against, so we can only use the monthly default value.
        return target.get_value()



import json

from crispy_forms.bootstrap import StrictButton
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.forms import CharField, JSONField
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django_modals.forms import ModelCrispyForm
from django_modals.modals import ModelFormModal
from django_modals.widgets.select2 import Select2

from advanced_report_builder.field_types import FieldTypes
from advanced_report_builder.field_utils import ReportBuilderFieldUtils
from advanced_report_builder.globals import DATE_FIELDS, NUMBER_FIELDS, LINK_COLUMNS, COLOUR_COLUMNS
from advanced_report_builder.models import ReportQuery, ReportType
from advanced_report_builder.utils import get_report_builder_class


class QueryBuilderModelForm(ModelCrispyForm):
    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        return StrictButton(button_text, onclick=f'save_modal_{ self.form_id }()', css_class=css_class, **kwargs)


class QueryBuilderModalBaseMixin(ReportBuilderFieldUtils):

    @staticmethod
    def get_report_builder_class(report_type_id):
        if not report_type_id:
            return None, None

        report_type = get_object_or_404(ReportType, pk=report_type_id)
        base_model = report_type.content_type.model_class()
        report_builder_class = get_report_builder_class(model=base_model,
                                                        report_type=report_type)
        return report_builder_class, base_model

    def get_query_builder_report_type_field(self, report_type_id):

        report_builder_class, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        if report_builder_class is None:
            # noinspection PyUnresolvedReferences
            return self.command_response()
        query_builder_filters = []
        self._get_query_builder_fields(base_model=base_model,
                                       query_builder_filters=query_builder_filters,
                                       report_builder_class=report_builder_class)

        return query_builder_filters

    def _get_query_builder_fields(self, base_model, query_builder_filters, report_builder_class, prefix='',
                                  title_prefix='', previous_base_model=None):
        field_types = FieldTypes()

        for report_builder_field in report_builder_class.fields:
            if (not isinstance(report_builder_field, str) or
                    report_builder_field not in report_builder_class.exclude_search_fields):
                django_field, _, columns, _ = self.get_field_details(base_model=base_model,
                                                                     field=report_builder_field,
                                                                     report_builder_class=report_builder_class)
                for column in columns:
                    field_types.get_filter(query_builder_filters=query_builder_filters,
                                           django_field=django_field,
                                           field=column.column_name,
                                           title=title_prefix + column.title,
                                           column=column,
                                           prefix=prefix)
        for include_field, include in report_builder_class.includes.items():
            app_label, model, report_builder_fields_str = include['model'].split('.')
            new_model = apps.get_model(app_label, model)
            new_report_builder_class = get_report_builder_class(model=new_model,
                                                                class_name=report_builder_fields_str)

            if new_model != previous_base_model:
                _foreign_key = getattr(base_model, include_field, None)

                if hasattr(_foreign_key, 'field'):
                    add_null_field = _foreign_key.field.null
                else:
                    add_null_field = True

                if add_null_field:
                    field_types.get_foreign_key_null_field(query_builder_filters=query_builder_filters,
                                                           field=prefix + include_field,
                                                           title=title_prefix + include['title'])
                if isinstance(new_model(), AbstractUser):
                    field_types.get_abstract_user_field(query_builder_filters=query_builder_filters,
                                                        field=prefix + include_field,
                                                        title=title_prefix + include['title'])

                self._get_query_builder_fields(base_model=new_model,
                                               query_builder_filters=query_builder_filters,
                                               report_builder_class=new_report_builder_class,
                                               prefix=f"{prefix}{include_field}__",
                                               title_prefix=f"{include['title']} --> ",
                                               previous_base_model=base_model)


class QueryBuilderModalBase(QueryBuilderModalBaseMixin, ModelFormModal):
    base_form = QueryBuilderModelForm
    size = 'xl'

    def __init__(self, *args, **kwargs):
        self.report_query = None
        self.show_query_name = False
        super().__init__(*args, **kwargs)

    def ajax_get_query_builder_fields(self, **kwargs):
        report_type_id = kwargs['report_type']

        field_auto_id = kwargs['field_auto_id']
        if report_type_id:
            query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)

            return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))
        else:
            return self.command_response()

    def add_query_data(self, form, include_extra_query=False):
        form.fields['query_data'] = JSONField(required=False, label='Filter')

        if include_extra_query:
            form.fields['extra_query_data'] = JSONField(required=False, label='Numerator filter')

        if self.object.id:
            query_id = self.slug.get('query_id')
            if query_id:
                self.report_query = get_object_or_404(ReportQuery, id=query_id)
            else:
                self.report_query = self.object.reportquery_set.first()

            if self.report_query:
                self.show_query_name = self.object.reportquery_set.count() > 1
                if self.show_query_name:
                    form.fields['query_name'] = CharField(required=True, initial=self.report_query.name)

                form.fields['query_data'].initial = self.report_query.query
                if include_extra_query:
                    form.fields['extra_query_data'].initial = self.report_query.extra_query

    @staticmethod
    def _is_search_match(search_string, title):
        if search_string is None:
            return True
        return search_string.lower() in title.lower()

    def _get_fields(self, base_model, fields, report_builder_class, tables=None,
                    prefix='', title_prefix='', title=None, colour=None,
                    previous_base_model=None, selected_field_id=None, for_select2=False,
                    pivot_fields=None, allow_annotations_fields=True, field_types=None,
                    column_types=None, search_string=None, show_order_by_fields=False, extra_fields=None):
        if title is None:
            title = report_builder_class.title
        if colour is None:
            colour = report_builder_class.colour

        if tables is not None:
            tables.append({'name': title,
                           'colour': colour})

        report_builder_class_fields = report_builder_class.fields

        if extra_fields:
            report_builder_class_fields += extra_fields

        for report_builder_field in report_builder_class_fields:

            if (not isinstance(report_builder_field, str) or
                    report_builder_field not in report_builder_class.exclude_display_fields or
                    (show_order_by_fields and report_builder_field in report_builder_class.order_by_fields)):
                django_field, col_type_override, columns, _ = self.get_field_details(
                    base_model=base_model,
                    field=report_builder_field,
                    report_builder_class=report_builder_class)
                for column in columns:
                    if ((field_types is None and column_types is None) or
                            (field_types is not None and isinstance(django_field, field_types)) or
                            (column_types is not None and isinstance(col_type_override, column_types)) or
                            (allow_annotations_fields and column.annotations)):
                        full_id = prefix + column.column_name
                        if selected_field_id is None or selected_field_id == full_id:

                            if column.title == '':
                                full_title = title_prefix + col_type_override.title_from_name(column.column_name)
                            else:
                                full_title = title_prefix + column.title
                            if self._is_search_match(search_string=search_string, title=full_title):
                                if for_select2:
                                    fields.append({'id': full_id,
                                                   'text': full_title})
                                else:
                                    fields.append({'field': full_id,
                                                   'label': full_title,
                                                   'colour': colour})

        if not for_select2 and pivot_fields is not None:
            for pivot_code, pivot_field in report_builder_class.pivot_fields.items():
                full_id = prefix + pivot_code
                full_title = title_prefix + pivot_field['title']
                if self._is_search_match(search_string=search_string, title=full_title):
                    pivot_fields.append({'field': full_id,
                                         'label': title_prefix + pivot_field['title'],
                                         'colour': colour})

        for include_field, include in report_builder_class.includes.items():
            app_label, model, report_builder_fields_str = include['model'].split('.')

            new_model = apps.get_model(app_label, model)
            if new_model != previous_base_model:
                new_report_builder_class = get_report_builder_class(model=new_model,
                                                                    class_name=report_builder_fields_str)
                self._get_fields(base_model=new_model,
                                 fields=fields,
                                 report_builder_class=new_report_builder_class,
                                 tables=tables,
                                 prefix=f"{prefix}{include_field}__",
                                 title_prefix=f"{title_prefix}{include['title']} -> ",
                                 title=include.get('title'),
                                 colour=include.get('colour'),
                                 previous_base_model=base_model,
                                 selected_field_id=selected_field_id,
                                 for_select2=for_select2,
                                 pivot_fields=pivot_fields,
                                 allow_annotations_fields=allow_annotations_fields,
                                 field_types=field_types,
                                 column_types=column_types,
                                 search_string=search_string,
                                 show_order_by_fields=show_order_by_fields)

    def ajax_get_fields(self, **kwargs):
        report_type_id = kwargs['report_type']
        report_builder_class, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        fields = []
        tables = []
        self._get_fields(base_model=base_model,
                         fields=fields,
                         tables=tables,
                         report_builder_class=report_builder_class,
                         field_types=NUMBER_FIELDS,
                         extra_fields=report_builder_class.extra_chart_field)
        return self.command_response('report_fields', data=json.dumps({'fields': fields, 'tables': tables}))

    def _get_date_fields(self, base_model, fields, report_builder_class, selected_field_id=None, search_string=None):
        return self._get_fields(base_model=base_model,
                                fields=fields,
                                report_builder_class=report_builder_class,
                                selected_field_id=selected_field_id,
                                field_types=DATE_FIELDS,
                                for_select2=True,
                                allow_annotations_fields=False,
                                search_string=search_string)

    def _get_number_fields(self, base_model, fields, report_builder_class, selected_field_id=None, search_string=None):
        return self._get_fields(base_model=base_model,
                                fields=fields,
                                report_builder_class=report_builder_class,
                                selected_field_id=selected_field_id,
                                field_types=NUMBER_FIELDS,
                                for_select2=True,
                                search_string=search_string
                                )

    def _get_column_link_fields(self, base_model, fields, report_builder_class,
                                selected_field_id=None, search_string=None):
        return self._get_fields(base_model=base_model,
                                fields=fields,
                                report_builder_class=report_builder_class,
                                selected_field_id=selected_field_id,
                                column_types=LINK_COLUMNS,
                                for_select2=True,
                                search_string=search_string,
                                allow_annotations_fields=False
                                )

    def _get_colour_fields(self, base_model, fields, report_builder_class,
                           selected_field_id=None, search_string=None):
        return self._get_fields(base_model=base_model,
                                fields=fields,
                                report_builder_class=report_builder_class,
                                selected_field_id=selected_field_id,
                                column_types=COLOUR_COLUMNS,
                                for_select2=True,
                                search_string=search_string,
                                allow_annotations_fields=False
                                )

    def get_fields_for_select2(self, field_type, report_type, search_string):
        fields = []
        if report_type != '':
            report_builder_fields, base_model = self.get_report_builder_class(report_type_id=report_type)
            fields = []
            if field_type == 'date':
                self._get_date_fields(base_model=base_model,
                                      fields=fields,
                                      report_builder_class=report_builder_fields,
                                      search_string=search_string)
            elif field_type == 'number':
                self._get_number_fields(base_model=base_model,
                                        fields=fields,
                                        report_builder_class=report_builder_fields,
                                        search_string=search_string)
            elif field_type == 'link':
                self._get_column_link_fields(base_model=base_model,
                                             fields=fields,
                                             report_builder_class=report_builder_fields,
                                             search_string=search_string)
            elif field_type == 'colour':
                self._get_colour_fields(base_model=base_model,
                                        fields=fields,
                                        report_builder_class=report_builder_fields,
                                        search_string=search_string)
            elif field_type == 'all':
                self._get_fields(base_model=base_model,
                                 fields=fields,
                                 report_builder_class=report_builder_fields,
                                 for_select2=True,
                                 search_string=search_string)
            elif field_type == 'order':
                self._get_fields(base_model=base_model,
                                 fields=fields,
                                 report_builder_class=report_builder_fields,
                                 for_select2=True,
                                 search_string=search_string,
                                 show_order_by_fields=True)

        return JsonResponse({'results': fields})

    def setup_field(self, field_type, form, field_name, selected_field_id, report_type):
        _fields = []
        if selected_field_id:
            form.fields[field_name].initial = selected_field_id
            base_model = report_type.content_type.model_class()
            report_builder_class = get_report_builder_class(model=base_model,
                                                            report_type=report_type)
            if field_type == 'date':
                self._get_date_fields(base_model=base_model,
                                      fields=_fields,
                                      report_builder_class=report_builder_class,
                                      selected_field_id=selected_field_id)
            elif field_type == 'link':
                self._get_column_link_fields(base_model=base_model,
                                             fields=_fields,
                                             report_builder_class=report_builder_class,
                                             selected_field_id=selected_field_id)
            elif field_type == 'number':
                self._get_number_fields(base_model=base_model,
                                        fields=_fields,
                                        report_builder_class=report_builder_class,
                                        selected_field_id=selected_field_id)
            elif field_type == 'colour':
                self._get_colour_fields(base_model=base_model,
                                        fields=_fields,
                                        report_builder_class=report_builder_class,
                                        selected_field_id=selected_field_id)
            elif field_type == 'all':
                self._get_fields(base_model=base_model,
                                 fields=_fields,
                                 report_builder_class=report_builder_class,
                                 selected_field_id=selected_field_id,
                                 for_select2=True)
            elif field_type == 'order':
                self._get_fields(base_model=base_model,
                                 fields=_fields,
                                 report_builder_class=report_builder_class,
                                 selected_field_id=selected_field_id,
                                 for_select2=True,
                                 show_order_by_fields=True)

        form.fields[field_name].widget = Select2(attrs={'ajax': True})
        form.fields[field_name].widget.select_data = _fields

    def form_valid(self, form):
        org_id = self.object.id if hasattr(self, 'object') else None
        chart_report = form.save()

        extra_query_data = form.cleaned_data.get('extra_query_data')
        if not self.report_query and (form.cleaned_data['query_data'] or extra_query_data):
            ReportQuery(query=form.cleaned_data['query_data'],
                        extra_query=extra_query_data,
                        report=chart_report).save()
        elif form.cleaned_data['query_data'] or extra_query_data:
            self.report_query.extra_query = extra_query_data
            self.report_query.query = form.cleaned_data['query_data']
            if self.show_query_name:
                self.report_query.name = form.cleaned_data['query_name']
            self.report_query.save()
        elif self.report_query:
            self.report_query.delete()
        url_name = getattr(settings, 'REPORT_BUILDER_DETAIL_URL_NAME', '')
        if org_id is None and url_name:
            url = reverse(url_name, kwargs={'slug': chart_report.slug})
            return self.command_response('redirect', url=url)
        else:
            return self.command_response('reload')



# Generated by Django 3.2.7 on 2022-04-04 11:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0007_auto_20220322_1939'),
    ]

    operations = [
        migrations.AddField(
            model_name='target',
            name='default_percentage',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='target',
            name='colour',
            field=models.CharField(blank=True, help_text='The colour when it gets displayed on a report', max_length=10, null=True),
        ),
        migrations.AlterField(
            model_name='target',
            name='default_value',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.2.7 on 2023-04-28 15:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0008_auto_20220404_1144'),
    ]

    operations = [
        migrations.AddField(
            model_name='singlevaluereport',
            name='average_end_period',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='singlevaluereport',
            name='average_scale',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Year'), (2, 'Quarter'), (3, 'Month'), (4, 'Week'), (5, 'Day')], null=True),
        ),
        migrations.AddField(
            model_name='singlevaluereport',
            name='average_start_period',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='barchartreport',
            name='axis_value_type',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Sum'), (2, 'Maximum'), (3, 'Minimum'), (4, 'Count'), (5, 'Average Sum from Count')], default=4, null=True),
        ),
        migrations.AlterField(
            model_name='funnelchartreport',
            name='axis_value_type',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Sum'), (2, 'Maximum'), (3, 'Minimum'), (4, 'Count'), (5, 'Average Sum from Count')], default=4, null=True),
        ),
        migrations.AlterField(
            model_name='linechartreport',
            name='axis_value_type',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Sum'), (2, 'Maximum'), (3, 'Minimum'), (4, 'Count'), (5, 'Average Sum from Count')], default=4, null=True),
        ),
        migrations.AlterField(
            model_name='piechartreport',
            name='axis_value_type',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Sum'), (2, 'Maximum'), (3, 'Minimum'), (4, 'Count'), (5, 'Average Sum from Count')], default=4, null=True),
        ),
        migrations.AlterField(
            model_name='singlevaluereport',
            name='single_value_type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Count'), (2, 'Sum'), (3, 'Count & Sum'), (4, 'Percent'), (5, 'Percent from Count'), (6, 'Average Sum from Count'), (7, 'Average Sum over Time')], default=1),
        ),
    ]



from django_datatables.columns import ColumnNameError
from django_datatables.datatables import ColumnInitialisor

from advanced_report_builder.exceptions import ReportError


class ReportBuilderFieldUtils:
    column_initialisor_cls = ColumnInitialisor
    def get_field_details(self, base_model, field, report_builder_class, table=None, field_attr=None):

        if field_attr is None:
            field_attr = {}
        if isinstance(field, str) and field in report_builder_class.field_classes:
            field = report_builder_class.field_classes[field]

        path = field
        original_column_initialisor = self.column_initialisor_cls(start_model=base_model,
                                                                  path=field,
                                                                  table=table,
                                                                  **field_attr)

        try:
            columns = original_column_initialisor.get_columns()
        except ColumnNameError as e:
            raise ReportError(e)
        django_field = original_column_initialisor.django_field
        col_type_override = None
        if columns:
            col_type_override = columns[0]
            if isinstance(field, str) and '__' in field and col_type_override.model_path == '':
                # I think there is a bug in datatables where the model path isn't always set! hence the following code
                field_parts = field.split('__')[:-1]
                model_path = '__'.join(field_parts) + '__'
                col_type_override.model_path = model_path
                col_type_override.field = col_type_override.field  # actions the setter
                path = model_path

        if django_field is None and columns:
            col_type_override = columns[0]
            if isinstance(col_type_override.field, str):
                if isinstance(field, str):
                    path_parts = field.split('__')[:-1]
                    path_parts.append(col_type_override.field.split('__')[-1])
                    path = '__'.join(path_parts)
                    column_initialisor = self.column_initialisor_cls(start_model=base_model, path=path, table=table,
                                                                     **field_attr)
                    column_initialisor.get_columns()
                    django_field = column_initialisor.django_field
                else:
                    column_initialisor = self.column_initialisor_cls(start_model=base_model,
                                                                     path=col_type_override.field,
                                                                     table=table,
                                                                     **field_attr)
                    column_initialisor.get_columns()
                    django_field = column_initialisor.django_field

        return django_field, col_type_override, columns, path



# Generated by Django 3.2.7 on 2022-02-16 09:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0003_auto_20220215_1219'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('output_type', models.CharField(blank=True, max_length=200, null=True)),
                ('view_name', models.CharField(max_length=200)),
                ('settings', models.JSONField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
    ]
